# Core Tutorial

## Questions

- How is the Kitchen Sink layed out, what differenciates atoms, molecules and organisms
- How often will I be needing to use more than just the directory I've set up the project in? 
	- If it is frequent, then will it be for primarily reading source code or making changes outside of that project directory?

## Naming Conventions

Model
: Item in app/model/Item.java

Controller
: Items in app/controllers/Items.java

Front-end Controller
: ItemFront in app/controllers/ItemFront.java

## Workspaces

- When generating the project via `core create ~/sites/new_project`, if the option to generate an IDE workspace is selected, then (using VS-Code as an example) in VS-Code, the workspace will contain many other folders
- This is due the modular nature of Core
- Hierarchy: Play Framework -> Core -> Mantle -> new_project
- TODO: I'll need to develop a way that Vim can use the workspace files

## Kitchen Sink

- The Kitchen Sink is a collection of pre-made styling components (with a web editor)
- Accessible via project.dev.enigma.lan/admin/kitchen-sink
- Web interface lets you see and edit the components styling
- Each attribute is defined by one of three modes:

Inherit
: Default value of component by whoever made the component

Styled
: Preferred option
: Opens a menu where you're able to change the attributes
: Some limitations

Manual
: Only intended for HIGHLY SPECIFIC overrides
: Set attribute to any valid CSS value

## Core/Play commands

Creating a new project
: core create ~/sites/new_project

Restarting site
: cd ~/sites/project
: core again (this is an alias for core end + core begin)
: core again --clean (this will wipe local caches)

Importing database to site
: core end
: core db:import --source=/where/ever/the/db/is/stored.zip
: core again

Creating core template TODO
: core tp

After creating new Javascript files (ANY)
: core js:update
: core js-update (this one seems to work better)

## Site Logs

system.out
: all Core related logs

system.error.out
: error severity logs

system.warning.out
: warning severity logs

system.info.out
: info severity logs

npm-out.log
: NPM + module logs (WebPack and BrowserSync)

## Styling and Core SASS

- The styling is generated via Core SASS and not just regular CSS
- The Core SASS module, recursively traverses the project and gathers all of the .scss files
- Once this has been done, it maintains the hierarchy of the site when combining all of the .scss files into one style.css file
- This happens for settings, framework, module styles and other styles.

## Custom Content Model

- Creating a Page Content Object (content item model) is done via core tp option 4
- The auto-generated class for this may not have the `@Override` annotation on `setupParams()`, make sure it does.
- In the guide this extended off of `GenericContent`

## Configurable Foundation Content Objects

- The `AlertContent` create previously worked off of the `GenericContent` class
- Another option is `FoundationConfigurableContent`, which can be configured via YAML
	- Useful for assigning arbitrary metadata-esque configuration data to a content object

## Definitions

Messages
: Evil
: Used to change the default messages on the content objects
: Make sure to finish and update these at the end so that the client is happy but you are not going crazy missing your content item
Slot
: Collection of content items 
: Added to a page in one go

Static Slot
: Rigid in organisation of content items
: Using one gives ALL content items associated
: Content Items cannot be moved or deleted

Flexible Slot
: Defines which content items are able to be used on a page
: None of the content items are added initially
: Can add individual items as needed

Content Object
: Have a controller and view to make them work
: Can be added to a slot (as Content Item) 
: `@ContentConfig`
	: Defines the options that we see when adding this to a CMS template slot
	: This is where the CMS user setting up the template can configure the component for that template slot
	: Options are set at the `CMS Template Level`
: `@KRUDFields`
	: Same as other models (see Pizza)
	: No separate list view so no need for `column` or `list`, `edit` is enough
	: Any editable fields I want to add to the CMS go here
: `@KRUDDialog`
	: Configuration of the dialog that appears when editing the content item in the CMS page editor
: `extends GenericContent`
	: Brings in functionality and fields which make this model work as a content item
: `public void setupParams()`
	: Variables that I wish to be present in the view go here
	: `params` Map argument should have fields put into it
	: Similar to the way that you add variables to a front controller's `@ControllerView` method via `PageUtils.addRenderArgs` except here it is handled in the model 
: `public String getSearchContent()`
	: Allows you to define what gets provided to the search index of the content item
Content Item
: Instance of a configurable content object
`@Transient`
: A transient field will stop Core from creating and using a database field for the speciried field.
: It will essentially become an 'empty' field, which can be given a getter method to returnan arbitrary value.
`@Lob`
: ensures that the database field will allow a large amount of data by default (default is 255 char limit)
`@HTML`
: makes the field display as a rich HTML text editor in the CMS
`@Audited`
: This is used on models
: This annotation activates Hibernate Envers functionality
	: Creates a new audit table for the model e.g. pizza_aud
	: Every time that the model is modified, a new copy will be added to the above table and given a revision
	: When an audited model is requested, it will return the most recent audit record, not the most recent version in the table
`@NotAudited`
: This is used on fields within a model
: This will stop that value from being audited/interacting with the Workflow system
: Used for things like stock count, values which don't need approval for or to be published
: Creates a non-audited relationship between audited parent and non-audited child.
: If you add a child to the parent, the change will be effective immediately without publishing the parent.
: If you update a child already attached to the parent, the change will be effective immediately without publishing the parent.
`@Audited(targetAuditMode=RelationTargetAuditMode.NOT_AUDITED)`
: By using the `@Audited` tag, Hibernate will create a join table between the parent and child, which is then audited from the perspective of the parent
	: For example, it will map a revision of the parent to the id of the child, if the revision of the parent changes, it will map the new revision of the parent to the existing id of the child, maintaining the audited relationship.
: If you add a a child to the parent, the change will only occur once the parent has been re-published
: If you update a child, the change will be effctive immediately without publishing the parent
: Good example of this approach is by attaching assets to a product.
	: Assets are non-audited because theycan be used in multiple places and updating he image in an asset, at the asset level, will be reflected immediately.
	: However, if you change an asset that's attached to a product, you might want to forece a re-publish before the asset changes go live
`@Workflow`
: This is used on models
: This tells Core that the model is part of the Workflow system and does the following:
	: Store the revision number of each model against each state:
		: pizza id: 1, revision: 20 = Published
		: pizza id: 1, revision: 22 = Draft
	: Returns the most appropriate revision for the state required in the workflow
KRUDFields
: KRUD system automatically provides controllers and views for most object manipulation in Core, KRUDFields are a way to control what the various views show
: list
	: Fields that are made available to a listing view
: column
	: Fields that will be displayed as columns in a listing view
	: Anything included here must also exist in `list`
: edit
	: Fields that are displayed in an editing view
: include
	: Fields that are included when this model is provided to another model
	: Like a child relationship
	: include on a child means that it will include that data on anything above it
: tab: 
	: this prefix to a value results in a tab for those items to live in
: nosort: 
	: This prefix stops a value from being sortable, good for when using a `@Transient`
Requester:
: TODO
`CoreWorkflowQuery`
: Same as Hibernate's `.find()` method, but will take into account Workflow states
: `.matchrenderMode()`
	: Method for differanciating between front-end and CMS controllers
	: If a front-end controller is found then it works as `.publishedMode()`
	: If a CMS controller is found then it works as `.draftMode()`
	: If the function is called from AJAX (such as `getPizzas_JSON()`) then it defaults to `.draftMode()`
:`draftMode()`
	: Modifies the underlying query to return ONLY DRAFT REVISIONS 
:`publishedMode()`
	: Modifies the underlying query to return ONLY PUBLISHED REVISIONS
`saveItem(Object obj)`
: A method of the above style within a controller, will automatically get called on save allow persistence within the Java persistence lifecycle
: For example: `saveItem(Pizza pizza)` would update the Pizza model and update the database
: You can run `obj._save()` in the function to persist the object, then execute some functions after the save, but that requires returning true from `saveItem(Object object)` as returning false means that the object still needs to be saved.
	
## Custom Serialisers

- When sending JSOn data down to the web application (typically via AJAX calls), a serialiser needs to be used to include the data we want to send back.
- By default, you'll use `new JSONSerializer().include("*")` to add EVERYTHING to the JSON blob.
	- This can be SLOW.
- Instead, you can use a custom serialiser with a `doTransform(Object object)` method to transform the data into something much smaller and consise.
- The constructor for these accepts a mode, which can be used to specify HOW the data is transformed and for what purpose.
	- Data for editing can be different to data for viewing.
- Make sure that in this method, you check that the object being passed in is of the right type, if you have pizza and drinks separate, trying to access a drinks topping would cause breaking errors.
- You can either use `.transform(new CustomSerialiser(mode), object)` on the `new JSONSerialiser()` after the `include()` and `exclude()`
	- It seems that the way to do this isn't the above but instead: `KRUDTransformer.addTransformerModes(newSerialiser, "list")` to add all the serialisers with a `list` mode to our serialiser. This will take into account extensions to our custom transfomer, so if the pizza transformer linked to a toppings transformer, then both would be initalised.
- It is possible to add a custom transformer to a class via CMS by adding `@For(Pizza.class)`.
	- With the default modes being: list, edit and include.
	- Using the `@For(...)` annotation means:
		- When no type is specified, any serialisation that happens on that object, it will assume that it can do the 3 defaults
			- For example, in the Pizza Shop tutorial, I could remove the check in the serialiser for whether the mode is list and instead just put the tag at the beginning, specifying the modes available
		- When type(s) is/are specified, the transformer is available to be used for those modes implicityly.
	- This can catch you out as, although the annotation specifies what you can do, it can be more unclear rather than using `KRUDTransformer.addTransformerModes(...,...)`

## Core Workflow Functionality

- Core Workflow (workflow from now on), is a workflow systek in which an object has a status, this status can be changed and doing so changes the output of the object.
- This allows for site administrators to make changes without them being visible on the front-end immediately, instead they need to go through a approval and publishing process.
- In the tutorial, we'd added `@Audited` and `@Workflow` to the Pizza model, but only `@Audited` to the Topping model
	- We don't deem it neccessary to publish changes made to Toppings
	- Because the parent is audited, the relationship to the child will also be audited, making it easier to manage Toppings in the workflow easier later if we need it
- When adding workflow functionality to already existing models, there is no mechanism to backwwards populate the audit table for existing data.
	- Dropping/truncating those tables will fix this upon the next reload
	- If `_aud` tables already exist for the tables to drop, drop those in addition, otherwise we'll have ghost values (scary pizzas)
	- When dropping tables, in HeidiSQL, there is a feature to select a number of rows and then `Copy as` -> `SQL FUNCTION` statements, whether that be inserts or updates
- We'll also need to add the correct permissions to `core-admin-roles.yml`
- Before implementing the Workflows, we'd used `Model.findAll()`, this returns the most up to date versions of our model.
	- We instead want this to be a `CoreWorkflowQuery` which will instead take into account the revisions and state of our model
	- `List<Pizza> pizzas = Pizza.findAll()` -> `List<Pizza> pizzas = new CoreWorkflowQuery("select p from Pizza p").publishedMode().getResultList()`
	- The `CoreWorkflowQuery` is the Core equivalent of .find() which can take a number of parameters

## Batch Jobs

- CoreBatchJobs are built on top of the Play Framework Async Job mechanism
- A job is run in an independent thread
- How this is displayed in the admin panel is controlled by the messages file
- `@CoreBatchJob.BatchParam(required = default false)` is used to add parameters to the Batch Job that can be accessed in the CMS
	- This value is then set in the DB which is where it is then read from, the parmeters set in code are just defaults
- `@CoreBatchJob.BatchHistory` is used for something! I'll ask a SR Dev if/when it becomes relevant!
- `doJob()`
	- Overriding this method is what allows us to add logic to our batch job
- `setProgress(increment, max value)` allows us to update the status of the job in the CMS panel
- `obj.save()` if we've updated a value, then we need to save the value using this method

## Java Persistence Lifecycle
- We're able to manipulate our model during the Java Persistence Lifecycle
- There are three ways of doing this:
	- In the controller via `saveItem(Object obj)` (triggered as a hook in KRUD so this will only overrite CMS saves)
	- Overriding `_save()` in the model:
		- TODO
	- Internal Callback Methods/JPA event hooks
		- If you need your function to be called on cascade, then using annotations before your functions is perfect
		- Most commonly used:
			- `@PrePersist`: before a new entity is persisted (added to EntityManager/saved for the first time) 
			- `@PostPersist`: after storing a new entity in the DB (during commit or flush) 
			- `@PreUpdate`: when an entity is identified as modified by the EntityManager 
			- `@PostUpdate`: after updating an entry in the database (during commit or flush) 
		- Relatively un-used
			- `@PostLoad`: after an entity has been retrieved from the database 
			- `@PreRemove`: when an entuty is marked for removal in the entity manager 
			- `@PostRemove`: after deleting an entuty from e database (during commit or flush)
		- An example would be:
			```
				@PreUpdate
				public void calculateBalance() {
					balance = initialBalance - totalSpend;
				}
			```
		- However, it is not guaranteed that an `@PreUpdate` callback is invoked when an entity is persisted and then modified in a single transaction
		- To fix this, you'll need to do something like the following:
			```
				@PreUpdate
				public void calculateBalance() {
					balance = initialBalance - totalSpend;
				}
				
				@PrePersist
				public void initialiseBalance() {
					balance = initialBalance - totalSpend;
				}

			```
			
## Core Sass
- Core SASS is a module for automatically pre-compiling SASS during the build process.
- It has bene extended to allow for some more modular behaviour to allow the use of SASS files outside the site directory without needing to specify imports.
- It works in the following 4 steps:
	- Front-end or build process requests
	- Compilation process gathers all the style files across the modules, recursively trawling them to identify any import statements in order to identify and locate any child files that also need gathering
	- Copies all the identified files to the temp workspace, adhering to the heirarchical overrive structure in play
		- e.g. a file in the site will take precedence over one in a module
	- Analyses the name and position of those files in the file structure to ascertain the priority order they get rendered into the output style.css/.scss file
- The heirarchy is as follows:
	- style.css
		- Settings
			- stylesheets/settings/
			- e.g. `stylesheets/settings/_settings.colours.scss`
			- e.g. `stylesheets/settings/_settings.fonts.scss`
		- Framework
			- stylesheets/framework/
			- e.g. `stylesheets/framework/_framework.scss`
		- Module Styles
			- stylesheets/
			- e.g. `stylesheets/_module.discord.scss`
		- Other Styles
			- stylesheets/
			- e.g. `stylesheets/_site.scss`
- Make sure to use the [BEM](https://getbem.com/naming/) naming convention for your styling!
- In the included example of `_pizzashop.scss` there are a number of key takeaways/features:
	- Mixins
		- Defining a mixin called `text-base` at the top of the file allows for it to be included later on in the stylesheet
		- Similar to a function, which can be called using `@include` elsewhere in the stylesheet
		- `text-base` could be a regular class here, but making it a mixin makes more sense as it's only intended for internal use throughout this stylesheet
	- Pseudo selectors
		- Using the `&` operator as a reference to the outer selector when creating class names
		- `&__heading` inside `.pizza` is the same as defining `.pizza` and `.pizza__heading` as distinct classes
		- No special functionality, just makes the stylesheet more readable and makes BEM easier to implement
	- Inuit
		- Making use of Inuit classes and variables like `.o-layout` to speed up the process by pulling in some of the framework's styling
		- TODO: look up more with Inuit or at least make some notes further down the line
	- BEM
		- The Block__Element--Modifier naming convention is what is used!
		- The strategy in the aforementioned file, is to define the `cont`/container class which will act as a generic container for the contnet
		- Further classes are used to style the heading and configure the 2 and 4 column grid
			- Which are Elements within the `cont` block
		- Then using completely separate `pizza` and `topping` classe to create the styling for the Blocks which occupy the sapaces within the grid's Elments.
		- This separation of concerns means that the styling of the grid and whatever is within the grid can be defined and modified independently of each other

## Core Assets

- We can add an asset to a model via the following:
```
	@ManyToOne // only 1 image can be assigned to a pizza object at one time
	@Asset(type="image") // mark the field as an asset attribute to be used by the CMS
	@Audited(targetAuditMode = RelationTargetAuditMode.NOT_AUDITED)
	public CoreAsset image;
```
- We'll need to add it to the KRUD fields as well, so that we can see a `thumbnail:image` in the `column`, and then adding in `image` to both the `list` and `edit` fields.
- To pass down the image down to the front end, we can use our custom transformer:
```
if(pizza.image != null){
     write("image", pizza.image.id);
}
```
- Since the HTML only requires the image ID for our `CoreAsset` object, we don't need to pass the entire thing down:
	- Groovy:  `<img src="${models.CoreThumbnail.url(pizza.image,	  'ffffff', 80,200, 'in', 'png').raw()}"/>`
	- Angular: `<img src="${models.CoreThumbnail.url('{{pizza.image}}','ffffff', 80,200, 'in', 'png').raw()}"/>`
	- TODO: https://enigma-interactive.atlassian.net/wiki/spaces/CORE/pages/21106229/CoreThumbnail+-+What+it+does+and+how+to+use+it
- If we want the application to accept now or custom file types when the admin user uploads a file, we need to update `conf/asset-types.mime.properties` and `conf/asset-types.properties`:
	- `asset-types.mime.properties` example:
```
# Document Types
doc=application/msword
docx=application/vnd.openxmlformats-officedocument.wordprocessingml.document
xls=application/vnd.ms-excel
xlsx=application/vnd.openxmlformats-officedocument.spreadsheetml.sheet
txt=text/plain
ppt=application/vnd.ms-powerpoint
pptx=application/vnd.openxmlformats-officedocument.presentationml.presentation
pdf=application/pdf
ppp=application/ppp
 
# Video Types
mp4=video/mp4
m4v=video/mp4
f4v=video/x-flv
flv=video/x-flv
ogv=video/ogg
3gp=video/3gpp
3g2=video/3gpp2
```
	- `asset-types.properties` example:
```
# Document Types
doc=document
docx=document
xls=document
xlsx=document
txt=document
text=document
ppt=document
pptx=document
pdf=document
ppp=document
 
# Video Types
mp4=video
m4v=video
f4v=video
flv=video
ogv=video
3gp=video
3g2=video
```
- It is also possible to restrict access to categories or sub-categories of assets
	- In the CMS, you can apply a restriction by editing a category in the Asset Manager and pick a restriction (this will be applied to every asset in the category)
	- It is also possible to setup custom restrictions using `CoreAssetRestrictor` and overriding the `allowRequest` method
		- Two examples can be found in `core_member`: 
			- `core.interfaces.CoreMemberFullRestrictor` 
				- Blocks all asset or thumbnail requests unless from a logged in member account
			- `core.interfaces.CoreMemberAssetRestrictor`
				- Allows access to all thumbnails, but blocks full asset requests unless from a logged in member account
				
## Link Providers				

- There are a number of features that warant implementing a link provider:
	- Link selector in the CMS to create a link to a Pizza selected from a list of published Pizzas
	- A breadcrumb on the single pizza and pizza list page
	- A navigation menu containing our pizzas
	- A physical sitemap HTML view to be displayed to the users
	- A sitemap XML file to be generated automatically for search engines
	- Short URLs for some pizzas e.g. /pizza/chicken -> /chicken
- The Sitemap XML feature is a core feature, disabled by default, but the `SiteFeatures.java` file, contains this along with other site features that can be enabled
	- `CoreFeature.enableFeature("core.sitemap");` uncommenting the line prior enables it!
- This/these features are best to just look at the code to understand, making notes won't be too helpful! (Look at the commit)
- TODO: post site rebuild, I'll need to do this properly

## OpenSearch Introduction

- OpenSearch (fork of ElasticSearch) is Enigmas favoured memory-reisdent search service for Core
- Checking if the service is running in WSL is just checking the service:
	- `systemctl status opensearch`
- Config for this is stored in `pizzashop/conf/local-wsl-MACHINE.conf`
- The port number/main config file can be found at `/etc/opensearch/opensearch.yml`
- Also make sure to check that `core_elasticsearch` is included in the `dependencies.yml` require array.
- From here on, the [old guide](https://enigma-interactive.atlassian.net/wiki/spaces/CORE/pages/21113999/Adding+ElasticSearch+to+the+Pizza+Shop) is used!
- In order to be able to query elasticsearch, you need a query template, stored in `/conf/es_templates/search.moustache`
- Then we need to add our model to the elasticsearch mappings file under `conf/elasticsearch-mappings.yml`

## Foundation Content Collections

## Issues found during tutorial

- For section 3 (CMS Templates and Pages), when the tutorial asks for you to login with dev@core.dev/dev on SITE_URL/admin, you'll get a 'Template execution error'
	- The fix for this is to go into the DB and update the `passwordChanged` value for the admin (only) user to NULL so that you are not prompted for a password change
- For section 3 (CMS Templates and Pages), under section 'Modifying the homepage content in CMS, subsection 'Publishing a page'; '2. Click the Instance Publish (on the ...' should be '2. Click the Instant Publish (on the ...'
- For section 5 (Modelling a Pizza), under [The Topping Model, and annotations](https://enigma-interactive.atlassian.net/wiki/spaces/ECD/pages/21299839/Core+Tutorial+-+5.+Modelling+a+Pizza+for+the+CMS#The-Topping-model%2C-and-annotations) it states that garlic bread may have toppings, this feels wrong to me. Not a major issue, but it felt blasphemous, especially since the author was also not sure about their statement.
- For section 5 (Modelling a Pizza), under [CMS-specific Annotations](https://enigma-interactive.atlassian.net/wiki/spaces/ECD/pages/21299839/Core+Tutorial+-+5.+Modelling+a+Pizza+for+the+CMS#CMS-specific-annotations), 'Saved the file, right? Run `core again` just in case it's a caching issue? And it's still just the title?' was a strange read as the answer is literally the line below, but the tone was very much 'It is working right??????' which didn't feel helpful. 'Saved the file, right? Run `core again` just in case it's a caching issue? And it's still just the title? Read ahead for the fix :)' would be nice! This is such a nit-pick but I did spend 10 min running `core again` and checking the DB to make sure that everything else was working fine.
- For section 6 (Create a new foundation content object), under [Setting up the default view](https://enigma-interactive.atlassian.net/wiki/spaces/ECD/pages/21299821/Core+Tutorial+-+6.+Create+a+new+foundation+content+object#Setting-up-the-default-view) it refers to 'when we ran `play tp` at the start...' rather than `core tp`. This is the same at the bottom of the page in `Wrapping Up`
- For section 6 (Create a new foundation content object), under [Adding it to a template and page](https://enigma-interactive.atlassian.net/wiki/spaces/ECD/pages/21299821/Core+Tutorial+-+6.+Create+a+new+foundation+content+object#Adding-it-to-a-template-and-page), the tutorial states that I should edit the `Site Base` template, then add my new `AlertContent` onto the `Homepage`. If you edit the `Site Base` template instead of the `Content Page` template, then when loading the `Homepage` and trying to add `AlertContent`, the system freaks out and just keeps loading after recieving an Internal Server Error via XHR. Doing the same change onto the `Content Page` template works without issue.
- For section 8C (Creating controller and view - Front end using VueJS) under [Vue App and Component](https://enigma-interactive.atlassian.net/wiki/spaces/ECD/pages/72057287/Core+Tutorial+-+8.+Creating+Controller+and+View+-+C.+Front-end+display+using+Vue#Vue-App-and-Component), when adding new .vue or VueJS files, it mentions running `core again --npm` in case NPM is not running. It should also mention running `core js:update` as after adding the VueJS files, the template would load but the Vue application wouldn't.
- For section 8C (Creating controller and view - Front end using VueJS), whenever it mentions `createApp({render: vueTemplateRenderFn("app");});`, the semi-colon flags up in VSCode as invalid syntax, then running `core js-update` returns an error on render.
- For section 9 (Serialise the data using a Transformer), under the very final 'Please note...', there is a link that states that there is more information about JSON serialisation, however, it links to a page about `How to use ng-init and ng-bind-html-unsafe`.
- For section 11 (Creating a batch job), under `Batch Job Example`, it specifies `First we will create a new package called 'core.jobs' in app`, this should read `First we will create a new package called 'core.jobs' in app/core/jobs`, it takes reading ahead to the contents of `RandomStockIncrement.java` to know where to put it. This might be on me as I have not done a lot of work in creating Java packages however and the text is immediately obvious to a more seasoned user.

## Versioning

Best way to tell what version of Core a site is using is just by checking the prod/pre prod CMS.


## Useful Links
- https://enigma-interactive.atlassian.net/wiki/spaces/ECD/pages/21299797/Core+Tutorial+-+The+Pizza+Shop
- https://www.playframework.com/documentation/1.5.x/cheatsheet/model
- https://enigma-interactive.atlassian.net/wiki/spaces/CORE/pages/21111746/Core+SASS
- https://enigma-interactive.atlassian.net/wiki/spaces/~lewisl/pages/21895249/Kitchen+Sink+Configurable+Components
- https://css-tricks.com/a-complete-guide-to-custom-properties/
- https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-multi-match-query.html
